<?php

/**
 * @package     Joomla.Plugin
 * @subpackage  System.accessibility
 *
 * @copyright   (C) 2020 Open Source Matters, Inc. <https://www.joomla.org>
 * @license     GNU General Public License version 2 or later; see LICENSE.txt
 */

namespace Joomla\Plugin\Content\IndexNow\Extension;

use Joomla\CMS\Application\CMSApplicationInterface;
use Joomla\CMS\Event\Model\AfterSaveEvent;
use Joomla\CMS\Event\Model\PrepareFormEvent;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\Table\Table;
use Joomla\Component\Categories\Administrator\Table\CategoryTable;
use Joomla\Component\Content\Administrator\Table\ArticleTable;
use Joomla\Event\DispatcherInterface;
use Joomla\Event\Priority;
use Joomla\Event\SubscriberInterface;
use Joomla\Plugin\Content\IndexNow\Client;
use Joomla\Plugin\Content\IndexNow\Config as IndexNowConfig;
use Joomla\Plugin\Content\IndexNow\Exception\ClientResponseException;
use Joomla\Plugin\Content\IndexNow\Provider\YandexProvider;
use Joomla\Plugin\Content\IndexNow\RateLimiter;
use Joomla\Plugin\Content\IndexNow\RegexMatcher;
use Joomla\Plugin\Content\IndexNow\Storage\StorageInterface;
use Joomla\Plugin\Content\IndexNow\UrlFetcher\UrlFetcherFactory;
use Psr\Http\Client\ClientExceptionInterface;
use Psr\Http\Client\ClientInterface;

// phpcs:disable PSR1.Files.SideEffects
\defined('_JEXEC') or die;
// phpcs:enable PSR1.Files.SideEffects

/**
 * System plugin to add additional accessibility features to the administrator interface.
 *
 * @since  4.0.0
 */
final class IndexNow extends CMSPlugin implements SubscriberInterface
{
    private const MODE_ASK = 'ask';
    private const MODE_AUTO = 'auto';

    private readonly RegexMatcher $regexMatcher;
    private readonly RateLimiter $rateLimiter;

    public function __construct(
        DispatcherInterface $dispatcher,
        private readonly ClientInterface $httpClient,
        readonly StorageInterface $storage,
        array $config = []
    ) {
        parent::__construct($dispatcher, $config);

        $this->rateLimiter = new RateLimiter(
            storage: $storage,
            minutes: (int) $this->params->get('min-submit-interval')
        );

        $this->regexMatcher = new RegexMatcher(
            explode(PHP_EOL, $this->params->get('exclude-patterns', ''))
        );
    }

    public function onContentPrepareForm(PrepareFormEvent $event): void
    {
        if ($this->params->get('mode') !== self::MODE_ASK) {
            return;
        }

        $wa = $this->getApplication()?->getDocument()->getWebAssetManager();

        $wa->registerAndUseScript(
            'plg_indexnow.submit',
            'plg_indexnow/indexnow/submit.js',
            [],
            ['type' => 'module'],
            ['core']
        )
        ->registerAndUseStyle('plg_indexnow.submit', 'plg_indexnow/indexnow/style.css');
    }

    /**
     * @param ArticleTable|CategoryTable $model
     *
     * @throws \JsonException
     * @throws ClientExceptionInterface
     * @throws \Exception
     *
     * @since  4.0.0
     */
    public function onContentAfterSave(AfterSaveEvent $event): void
    {
        $this->loadLanguage();

        if (!$event->getItem() instanceof ArticleTable && !$event->getItem() instanceof CategoryTable) {
            return;
        }

        if (! $this->checkToken($this->params->get('key'))) {
            return;
        }

        if (! $this->canProcessContent($event->getItem())) {
            return;
        }

        $fetcher = $event->getItem() instanceof ArticleTable
            ? UrlFetcherFactory::buildFromArticle($event->getItem())
            : UrlFetcherFactory::buildFromCategory($event->getItem());

        if (! $fetcher->isValid()) {
            return;
        }

        try {
            $this->regexMatcher->match($fetcher->url());
        } catch (\Throwable $e) {
            $this->notify(Text::sprintf(
                'PLG_CONTENT_INDEXNOW_IN_URL_EXCLUDED_BY_PATTERN_XY',
                $fetcher->url(),
                $e->getMessage()
            ), CMSApplicationInterface::MSG_WARNING);

            return;
        }

        if (! $this->rateLimiter->canProcess($fetcher->url())) {
            $this->notify(Text::sprintf(
                'PLG_CONTENT_INDEXNOW_IN_URL_SUBMITTED_X_MINUTES_AGO',
                (int)$this->rateLimiter->getElapsedMinutes($fetcher->url()),
                $this->rateLimiter->getRemainingMinutes($fetcher->url())
            ), CMSApplicationInterface::MSG_WARNING);

            return;
        }

        switch ($this->params->get('mode')) {
            case self::MODE_ASK:
                $this->displayManualTrigger($fetcher->url());
                break;
            case self::MODE_AUTO:
                $this->sendUrl($fetcher->url()) && $this->rateLimiter->store($fetcher->url());
                break;
            default: throw new \RuntimeException('Unsupported mode');
        }
    }

    /**
    * @throws ClientExceptionInterface
    * @throws \JsonException
    */
    public function onAjaxIndexnow(): void
    {
        if (! $this->checkToken($this->params->get('key'))) {
            return;
        }

        $url = $this->getApplication()->input->post->get('url', null, 'raw');

        if (! filter_var($url, FILTER_VALIDATE_URL)) {
            return;
        }

        $url = filter_var($url, FILTER_SANITIZE_URL);

        if ($this->sendUrl($url)) {
            $this->rateLimiter->store($url);
        }
    }

    public function checkToken(string $token): bool
    {
        if (trim($token) === '') {
            $this->notify(Text::sprintf('PLG_CONTENT_INDEXNOW_IN_ERR_NO_KEY_SET_X'));
            return false;
        }

        if (strlen($token) < 8 || strlen($token) > 128 || !preg_match('/^[a-z0-9-]+$/i', $token)) {
            $this->notify(Text::sprintf('PLG_CONTENT_INDEXNOW_IN_ERR_KEY_INVALID_X'));
            return false;
        }

        return true;
    }

    private function notify(string $message, string $type = CMSApplicationInterface::MSG_NOTICE): void
    {
        $this->getApplication()?->enqueueMessage(
            msg: $message,
            type: $type
        );
    }

    private function displayManualTrigger(string $url): void
    {
        $this->notify(
            Text::sprintf(
                'PLG_CONTENT_INDEXNOW_IN_NOTIFICATION_SENDABLE_X',
                $url,
                $url,
                Text::_('PLG_CONTENT_INDEXNOW_IN_NOTIFICATION_SEND_NOW')
            )
        );
    }

    /**
     * @throws ClientExceptionInterface
     * @throws \JsonException
     */
    private function sendUrl(string $url): bool
    {
        $config = new IndexNowConfig(token: $this->params->get('key'));
        $provider = new YandexProvider(config: $config);

        $client = new Client(
            provider: $provider,
            httpClient: $this->httpClient,
        );

        try {
            $client->sendToIndex($url);
        } catch (ClientResponseException $e) {
            $details = $e->getDetails();

            $this->notify(
                message: sprintf('%s ==> %s: %s', 'IndexNowPlugin', $e->getMessage(), $details['message'] ?? ''),
                type: CMSApplicationInterface::MSG_ERROR,
            );

            return false;
        }

        $this->notify(
            Text::sprintf('PLG_CONTENT_INDEXNOW_IN_NOTIFICATION_SENT_X', $url),
        );

        return true;
    }

    private function canProcessContent(Table $model): bool
    {
        $processContent = filter_var($this->params->get('support-com_content', false), FILTER_VALIDATE_BOOLEAN);
        $processCategory = filter_var($this->params->get('support-com_categories', false), FILTER_VALIDATE_BOOLEAN);;

        return match (true) {
            $processContent && $model instanceof ArticleTable,
                $processCategory && $model instanceof CategoryTable => true,
            default => false,
        };
    }

    public static function getSubscribedEvents(): array
    {
        return [
            'onContentAfterSave' => ['onContentAfterSave', Priority::NORMAL],
            'onAjaxIndexnow' => ['onAjaxIndexnow', Priority::NORMAL],
            'onContentPrepareForm' => ['onContentPrepareForm', Priority::NORMAL]
        ];
    }
}
