<?php

declare(strict_types=1);

namespace Joomla\Plugin\Content\IndexNow;

use Joomla\Plugin\Content\IndexNow\Storage\StorageInterface;

final class RateLimiter
{
    private readonly \DateTimeInterface $now;

    public function __construct(
        private readonly StorageInterface $storage,
        private readonly int $minutes,
    )
    {
        $this->now = new \DateTimeImmutable();
    }

    public function canProcess(string $key): bool
    {
        if ($this->minutes === 0) {
            return true;
        }

        if (($elapsedMinutes = $this->getElapsedMinutes($key)) === null) {
            return true;
        }

        return $elapsedMinutes > $this->minutes;
    }

    public function store(string $key): void
    {
        if ($this->minutes === 0) {
            return;
        }

        $this->storage->set(
            key: self::prepareKey($key),
            value: $this->now->format(\DateTimeInterface::RFC3339),
        );
    }

    public function getElapsedMinutes(string $key): ?int
    {
        if (($savedDate = $this->getSavedDate($key)) === null) {
            return null;
        }

        return $this->getTimeDiff($savedDate)->i;
    }

    public function getRemainingMinutes(string $key): int
    {
        $elapsed = $this->getElapsedMinutes($key);

        if($elapsed === null || $this->minutes < $elapsed) {
            return 0;
        }

        return $this->minutes - $elapsed;
    }

    private function getSavedDate(string $key): ?\DateTimeInterface
    {
        $date = \DateTimeImmutable::createFromFormat(
            format: \DateTimeInterface::RFC3339,
            datetime: $this->storage->get(self::prepareKey($key), '')
        );

        return $date ?: null;
    }

    private static function prepareKey(string $key): string
    {
        return hash('md5', $key);
    }

    private function getTimeDiff(\DateTimeInterface $savedDate): \DateInterval
    {
        return $this->now->diff($savedDate);
    }
}
